function [refinedline,linecosts] = disparityplausibilitynew2PROD_SGMcosts(leftimage,rightimage,row,bestdisps,dispsofpreviousrow,params)
%function refinedline = disparityplausibilitynew2PROD(leftimage,rightimage,line,bestdisps,D);
%Difference to disparityplausibilitynew2 : Trimmed for efficiency. No direct execution via F5 possible  
%Right and left image are raw values!
%Take the dominating disparities (and some additional) values found by hough voting and test them 
%for plausibility

[hi,wi] = size(leftimage);
windowlength = params.CTlength;
windowheight = params.CTheight;

maxHDvalue = params.P2;
D = params.numberOfDisparities;

lenind = floor(windowlength / 2);
heiind = floor(windowheight / 2);

disparitiestotest = definedisparitiestotest2(bestdisps,D,wi);   %Expand disparities

if (row - heiind <= 0 || row + heiind > hi)
    leftstripe = borderfunTB(leftimage,row,heiind); %stripe, currentrow, heind)
    rightstripe = borderfunTB(rightimage,row,heiind);     
else
    leftstripe = leftimage(row - heiind : row + heiind,:);
    rightstripe = rightimage(row - heiind : row + heiind,:);
end

%bigdata = zeros(wi,11); %gt, sadind, sadval, sadcrossind, sadcrossval, ...
refinedline = zeros(wi,1);
linecosts = ones((wi-D)*D,1) * maxHDvalue;

%Run from 1 to end. Use each column number and no padded image for easier calulation of
%disparities
for col = D + 1 : wi
   
    %Check, if column counter is at left or right border. If so, call
    %function which padds the stripe with replicated columns
    if (col >= lenind + 1 && col <= wi - lenind)
        currentleftblock = leftstripe(:,col - lenind : col + lenind);   %left cut out to match
    else
        currentleftblock = borderfunLR(leftstripe,col,lenind);
    end
    
    %Get features of left pixel centered at line,col
    censusvecL = census(currentleftblock);                           %Census with window size
    censusLRbest = Inf; censusLRbestind = 0;
        
    currentblock = ceil(col / D);
    previousdisp = dispsofpreviousrow(col); %Get the disp of current x of previous row
    
    disparitiestotestofthisblock = disparitiestotest(currentblock,:);   %Expanded disparities
    disparitiestotestofthisblock = insertdisparities(disparitiestotestofthisblock, previousdisp, D);    %Insert disp of previous row
    lendispstotest = sum(isfinite(disparitiestotestofthisblock));   %Check length
        
    for d = 1 : lendispstotest
       %targetcol = col - disparitiestotest(currentblock,d);     %col is current x
       targetcol = col - disparitiestotestofthisblock(d);
       kleinD = disparitiestotestofthisblock(d);
       if targetcol <= 0
           break;    %disps to test are ordered. If target column is lower than 1, all remaining disps do fail also
       end
       
       if(targetcol >= lenind + 1 && targetcol <= wi - lenind)
           currentrightblock = rightstripe(:,targetcol - lenind : targetcol + lenind);
       else
           currentrightblock = borderfunLR(rightstripe,targetcol,lenind);
       end
       
       censusvecR = census(currentrightblock);
       censusLR = hammingdistance(censusvecL, censusvecR);
       
       
       if censusLR < censusLRbest
           censusLRbest = censusLR;
           %censusLRbestind = disparitiestotest(currentblock,d);
           censusLRbestind = disparitiestotestofthisblock(d);
       end
       
       currentxd = (col - D - 1)*D + kleinD;
       linecosts(currentxd) = censusLR; 
       
    end
    
    refinedline(col) = censusLRbestind;    %Collect the best indices!
    
end % of main for


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%Subfunctions:

function block = borderfunLR(stripe,col,lenind)
    %Padd stripe at left and right border with replicated columns
    %for windowlength of e.g. 5 (lenind = 2), col is either 1, 2 or 449, 450
    
    [~,wiloc] = size(stripe);
    if(col <= lenind)
        demand = lenind - col + 1;  %left border, how many columns are required?
        block = padarray(stripe(:,1:col+lenind),[0 demand],'replicate','pre');  %Replicate the first column of stripe demand times
    else
        demand = col + lenind - wiloc;  %right border
        block = padarray(stripe(:,col-lenind : wiloc),[0 demand],'replicate','post');  %Replicate the first column of stripe demand times
    end
end

function stripeback = borderfunTB(image, currentrow, heiind)
    %Padd stripe at top and bottom border with replicated rows
    %for windowheight of e.g. 5 (heiind = 2), h is either 1, 2 or 374, 375
    
    [hiloc,~] = size(image);
    if(currentrow <= heiind)
        demand = heiind - currentrow + 1;
        stripeback = padarray(image(1 : currentrow + heiind,:),[demand 0],'replicate','pre');
    else
        demand = currentrow + heiind - hiloc;
        stripeback = padarray(image(currentrow - heiind : hiloc,:),[demand 0],'replicate','post');
    end
end


function censusvec = census(block)
    [hiloc,wiloc] = size(block);
    
    cenY = ceil(hiloc / 2); cenX = ceil(wiloc / 2);
    center = block(cenY,cenX);
    censusvec = zeros((hiloc * wiloc) - 1,1);
    counter = 1;
    for h = 1 : hiloc
        for w = 1 : wiloc
            if (block(h,w) >= center)
                censusvec(counter) = 1;
            end
            if h == cenY && w == cenX 
                censusvec(counter) = 0; %Set center to 0 as it does not count!
            else
                counter = counter + 1;
            end
        end
    end
end

function disparitiestotest = definedisparitiestotest2(bestdisps,D,wi)
    %Strategy: 1. Add +/-disprange to each dominant disparity per block. 
    %2. Copy neighboring disparitiis to test, i.e. copy from 2 to 1, from
    %1,3 to 2, from 2,4 to 3 ...
    %more at beginning und end...
    [numberofblocks, dispsperblock] = size(bestdisps);
    disprange = 2;
    if disprange ~= 2
        display(['Warning. Disparity expansion by only +-', num2str(disprange)]);
    end
    finaldispsperblock = (disprange*2+1) * dispsperblock;
    disparitiestotesttemp = NaN(numberofblocks,finaldispsperblock);
    disparitiestotest = NaN(numberofblocks,finaldispsperblock*3);  %Test 3 disparities per block within a range of +/-disprange
    
    for b = 1 : numberofblocks
        wstart = ((b - 1) * D) + 1;
        wend = min(b*D,wi);
        actualdisps = (wend - wstart + 1) - bestdisps(b,:); %Transform indices to actual disparity values
        
        for i = 1 : dispsperblock
            actdisp = actualdisps(i);
            actdisprange = actdisp - disprange : actdisp + disprange;   %Add +/-disprange to each disparitiy
            disparitiestotesttemp(b,(i-1)*(2*disprange+1)+1 : (i*(2*disprange+1))) = actdisprange;
        end
    end
    
    %Rejection of invalid 
    for b = 1 : numberofblocks
        for i = 1 : dispsperblock * (2*disprange+1)
            actdisp = disparitiestotesttemp(b,i);
            if actdisp <= 0 || actdisp > D
                disparitiestotesttemp(b,i) = NaN;   %Remove invalid disparities below 1 or above D
            end
        end
    end
    
    %Copy the disparities of a block to its neighbors
    disparitiestotest(1,1:finaldispsperblock) = disparitiestotesttemp(1,:); %Copy first two rows to first row 
    disparitiestotest(1,finaldispsperblock + 1 : finaldispsperblock * 2) = disparitiestotesttemp(2,:);
    for b = 2 : numberofblocks - 1
        disparitiestotest(b,1:finaldispsperblock) = disparitiestotesttemp(b-1,:);   %Copy pre, current and post to current row
        disparitiestotest(b,finaldispsperblock + 1 : 2 * finaldispsperblock) = disparitiestotesttemp(b,:);
        disparitiestotest(b,2*finaldispsperblock + 1 : 3 * finaldispsperblock) = disparitiestotesttemp(b+1,:);
    end
    disparitiestotest(numberofblocks,1:finaldispsperblock ) = disparitiestotesttemp(numberofblocks-1,:);    %Copy last two ros to last row
    disparitiestotest(numberofblocks,finaldispsperblock+1:finaldispsperblock*2) = disparitiestotesttemp(numberofblocks,:);
    
    %Removal of duplicates
    for b = 1 : numberofblocks
        actdisp = unique(disparitiestotest(b,:));
        disparitiestotest(b,:) = NaN;
        disparitiestotest(b,1:length(actdisp)) = actdisp;
    end
end

function res = insertdisparities(disparitiestotestofthisblock, previousdisp, D)
   %Receives a vector with a number of disparitiestotest. Possibly with
   %NaNs (but not necessarily). Receives a single value which must be
   %expanded, and inserted at the right place in
   %disparitiestotestofthisblock, without doubles and without unvalid data (negatives, above D)!
   %Puh!
   
   dispdelta = 10;  %add +/- dispdelta around dominting disparity
   
   disparitiestotestofthisblock(isnan(disparitiestotestofthisblock)) = Inf; %Replace NaNs with Inf
   %lenofdispstotest = sum(isfinite(disparitiestotestofthisblock));
   prevdispexp = previousdisp - dispdelta : previousdisp + dispdelta;
   
   prevdispexpvalid = prevdispexp(prevdispexp > 0); %Remove negative
   prevdispexpvalid = prevdispexpvalid(prevdispexpvalid < D);   %Remove values above D
   
   res = unique([disparitiestotestofthisblock, prevdispexpvalid]); %remove double entries
   res(isinf(res)) = NaN;   %Replace Inf values with NaN
   
   
    
end



end     %of main function
















