function disparity = rebuildDZSAD_refac(img1, img2, params)
display('rebuildDZSAD_refac running...');
%%%%%%%%%%%%%%%%%%%%%
%SGM

%CAUTION: Uses kind of clipping. The costs are aggregated only, when they are lower
%than P2. I.e. the cost volume must be build in a way, that the maximum
%cost value is P2. 

%Assumptions:
%[hi,wi] = [height, width] of both input images
%minDisparity = minD = 0;
%maxD = minD + numberOfDisparities = numberOfDisparities
%D = maxD - minD = maxD = numberOfDisparities
%minX1 = maxD = D
%maxX1 = width + min(minD, 0) = wi
%width1 = wi - numberOfDisparities
%INVALID_DISP = -1
%r = 4 = number of SGBM paths. looking at a pixel p(y,x), r=0 comes from left (y,x-1), r=1 from top left (y-1,x-1), r=2
%from top (y-1,x) and r=3 from top right (y-1, x+1)

%P1 = params.P1;
%P2 = params.P2;
%minDisparity = params.minDisparity;
numberOfDisparities = params.numberOfDisparities; % disparity range
D = numberOfDisparities;
img1 = double(img1);
img2 = double(img2);    

[height, width] = size(img1);
relwidth = width - D;

%Implements two versions, old and new:
%Old steps over all d=1...D values. New only over those, indicated as good from cost calculation via indexvolume. 
%Indexvolume has the same size as cost volume. For each x, there are D/2
%indices indicating good cost values. the rest D/2 values of that pixel are 0. 

useold = 1; 

costsready = 0;
Cmatrix = zeros(height, (width-D)*D);   %call cost claulation function later (line by line!)
indexvolume = zeros(height,(width-D)*D,'uint8');

% P1 = 8*1*params.SADWindowSize^2;
% P2 = 32*1*params.SADWindowSize^2;
%Overwrite P1, P2
P1 = 10;
P2 = 100;

params.P1 = 10;
params.P2 = 100;

%For each path, an own cost buffer:
Lr0n = zeros(relwidth, D);      %n = 0, i.e. current line
Lr1n = zeros(relwidth, D);      %Except for L0, all other Ls come from previous line, i.e. their values must be preserved!
Lr1n1 = zeros(relwidth, D);
Lr2n = zeros(relwidth, D);
Lr2n1 = zeros(relwidth, D);
Lr3n = zeros(relwidth, D);
Lr3n1 = zeros(relwidth, D);

minLr0n = zeros(1,relwidth); %All minimums of Lr0 of one line
minLr1n = zeros(1,relwidth);
minLr1n1 = zeros(1,relwidth);
minLr2n = zeros(1,relwidth);
minLr2n1 = zeros(1,relwidth);

minLr3n = zeros(1,relwidth);
minLr3n1 = zeros(1,relwidth);

disparity = ones(height, width) * -1;   %Unvalid disparities are set to -1

%main loop over each line:
for y = 0 : height - 1  %Rest of world index
    if costsready == 1
        %Costs already calculated, i.e. use preloaded Cmatrix:
        Cline = Cmatrix(y+1,:); %C contains all costs of all image pixels. Because we operate line by line, get here the pixel costs of the current line
        indexline = indexvolume(y+1,:);
    else
       if y == 0
           dispsofpreviousrow = ones(1,width) * D/2;
       else
           dispsofpreviousrow = disparity(y,:);
       end
       Cline = preselectionfeaturesfromvoting(img1, img2, y+1, dispsofpreviousrow, params); %single line of costs
       Cmatrix(y+1,:) = Cline;
       indexline = zeros(1,(width-D)*D);
    end
    
    Sline = ones(relwidth * D,1) * P2;    %Contains the aggregated pixel costs for all pixels and disparities in this row
    
    
    %If at first row, init the aggregation cost buffers for those paths, which come from y-1
    if y == 0
        Lr1n1(1,:) = 0;     %L1 comes from previous row, i.e. init L1n1 correctly at beginning
        minLr1n1(:) = 0;
        
        Lr2n1(1,:) = 0;     %If at first row, init Lr2n1. If not, Lr2n1 was set in the previous run!
        minLr2n1(:) = 0;    
        
        Lr3n1(1,:) = 0;     %L3 comes from previous row, next pixel, i.e. L3n1(x+1)
        minLr3n1(:) = 0;
    end
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    %L0, L1:        
    
    Crange = Cline(1 : D);  %Get the first D costs from Cline for the first pixel in the row
    indexrange = indexline(1:D);

if useold ~= 1    
    %Init L0, L1 buffers for x=1:
    Lr0n(1,1:D) = P2;       %For x = 1, init L0 (which normally comes from previous x in same row):
    Lr1n(1,1:D) = P2;
    Lr1n1(1,1:D) = P2;
    Sline(1:D) = P2 + P2;       %not necessary, because Sline is all initialized with P2
    
	for i = 1 : round(D/2)
        d = indexrange(i);  %Select d from index 
        if d > 0
            cpd = Crange(d);
            L0 = cpd - P2;  
            L1 = cpd - P2;     %Lr1n is not set and not read in any loop any more. Hence, set it here, otherwise it remains 0

            Lr0n(1,d) = L0;
            Lr1n(1,d) = L1;
            Lr1n1(1,d) = L1;
            Sline(d) = L0 + L1; 
        end
    end

else

%% L0, L1, first x in line. 
    minLr0n(1) = min(Lr0n(1,:));       %Get init for minLr0(1) at first column
    minLr1n1(1) = min(Lr1n1(1,:));
    for d = 1 : D
        cpd = Crange(d);
        if cpd >= P2
            L0 = P2;  
            L1 = P2;
        else
            L0 = cpd - P2;  
            L1 = cpd - P2;     %Lr1n is not set and not read in any loop any more. Hence, set it here, otherwise it remains 0
        end
        Lr0n(1,d) = L0;
        Lr1n(1,d) = L1;
        Lr1n1(1,d) = L1;
        Sline(d) = L0 + L1;
    end
    
    minLr0n(1) = min(Lr0n(1,:));       %Get init for minLr0(1) at first column
    minLr1n1(1) = min(Lr1n1(1,:));
end
%%

    minLr0n(1) = min(Lr0n(1,:));       %Get init for minLr0(1) at first column
    minLr1n1(1) = min(Lr1n1(1,:));
    %Don't start with 1, because L0 is calulated from x=1
    for x = 2 : relwidth     %0 .. width1, 0 ... wi - 
        %x indicates current pixel. x-1 is previous pixel
        
        Crange = Cline((x-1)*D + 1 : (x)*D);  %get D (the relevant) cost values for the current pixel
        indexrange = indexline((x-1)*D + 1 : x*D);
         
        delta0 = minLr0n(x-1) + P2;
        delta1 = minLr1n1(x-1) + P2;   %Previous row, previous pixel    

if useold ~= 1        
        Lr0n(x,:) = P2;
        Lr1n(x,:) = P2;
        for i = 1 : round(D/2)
            d = indexrange(i);  %indexrange may contain 0s
            if d > 0
                cpd = Crange(d);
                if d > 1 && d < D
                    %Regular case, d = 2 .. D - 1:
                    L0 = cpd + min(Lr0n(x-1,d), min(Lr0n(x-1,d+1)+P1, min(Lr0n(x-1,d-1)+P1, delta0))) - delta0;
                    L1 = cpd + min(Lr1n1(x-1,d), min(Lr1n1(x-1,d+1)+P1, min(Lr1n1(x-1,d-1)+P1, delta1))) - delta1;
                else
                    %d==1 || d==D
                    if d == 1
                       L0 = cpd + min(Lr0n(x-1,d), min(Lr0n(x-1,d+1)+P1, delta0)) - delta0; 
                       L1 = cpd + min(Lr1n1(x-1,d), min(Lr1n1(x-1,d+1)+P1, delta1)) - delta1; 
                    else
                       L0 = cpd + min(Lr0n(x-1,d), min(Lr0n(x-1,d-1)+P1, delta0)) - delta0;
                       L1 = cpd + min(Lr1n1(x-1,d), min(Lr1n1(x-1,d-1)+P1, delta1)) - delta1;
                    end
                end
                Lr0n(x,d) = L0;
                Lr1n(x,d) = L1;
                Sline(((x-1)*D) + d) = L0 + L1;
            end
        end
        
        minLr0n(x) = min(Lr0n(x,:));    %Minimum of D values
        minLr1n(x) = min(Lr1n(x,:));
else        
%% L0, L1 for all d = 1...D. 
      minLr0n(1) = min(Lr0n(1,:));       %Get init for minLr0(1) at first column
      minLr1n1(1) = min(Lr1n1(1,:));

        for d = 1 : D
            cpd = Crange(d);    %Get costs of current pixel at currently processed d
            
            if cpd >= P2
                L0 = P2;
                L1 = P2;
            else
                if d > 1 && d < D
                    %Regular case, d = 2 .. D - 1:
                    L0 = cpd + min(Lr0n(x-1,d), min(Lr0n(x-1,d+1)+P1, min(Lr0n(x-1,d-1)+P1, delta0))) - delta0;
                    L1 = cpd + min(Lr1n1(x-1,d), min(Lr1n1(x-1,d+1)+P1, min(Lr1n1(x-1,d-1)+P1, delta1))) - delta1;
                else
                    %d==1 || d==D
                    if d == 1
                       L0 = cpd + min(Lr0n(x-1,d), min(Lr0n(x-1,d+1)+P1, delta0)) - delta0; 
                       L1 = cpd + min(Lr1n1(x-1,d), min(Lr1n1(x-1,d+1)+P1, delta1)) - delta1; 
                    else
                       L0 = cpd + min(Lr0n(x-1,d), min(Lr0n(x-1,d-1)+P1, delta0)) - delta0 ;
                       L1 = cpd + min(Lr1n1(x-1,d), min(Lr1n1(x-1,d-1)+P1, delta1)) - delta1;
                    end
                end
            end

            Lr0n(x,d) = L0;
            Lr1n(x,d) = L1;
            Sline(((x-1)*D) + d) = L0 + L1;
        
        end
        
        %At end of one pixel. set minL0, 1, 2, 3 as minimum of Lr0, Lr1 .. in range over entire
        %d for a x, z.B. irgendwie so: minL0 = min(Lr0(1,:))
        minLr0n(x) = min(Lr0n(x,:));      %Minimum of D values
        minLr1n(x) = min(Lr1n(x,:));
%%
end     %of if useold ~= 1
        
    end     %of x = 2 : relwidth (Pixel loop for L0 and L1)
    
    %L0 done. No current line to previous shift necessary, because L0
    %doesn't need a previous row. It is calulated from the previous x in
    %the same row. A correct init is enough!
    
    Lr1n1 = Lr1n;       %Shift current L1 to previous
    minLr1n1 = minLr1n;
   
    
%%%%%%%%%%%%%%%%%%%%
    %L2:    
    
    for x = 1 : relwidth     %0 .. width1, 0 ... wi - 
        %x indicates current pixel. x-1 is previous pixel
        
        Crange = Cline((x-1)*D + 1 : (x)*D);  %get the relevant costs for the current pixel
        delta2 = minLr2n1(x) + P2;  %Same x, previous line!

if useold ~= 1        
        indexrange = indexline((x-1)*D + 1 : x*D);
        Lr2n(x,:) = P2;     %Init Lr2n because not all of its indices are touched any more
     
		for i = 1 : round(D/2)
            d = indexrange(i);
            if d > 0
                cpd = Crange(d);

                if d > 1 && d < D
                    %Regular case, d = 2 .. D - 1:
                    L2 = cpd + min(Lr2n1(x,d), min(Lr2n1(x,d+1)+P1, min(Lr2n1(x,d-1)+P1, delta2))) - delta2;
                else
                    %cases d==1 || d==D
                    if d == 1
                       L2 = cpd + min(Lr2n1(x,d), min(Lr2n1(x,d+1)+P1, delta2)) - delta2; 
                    else
                       L2 = cpd + min(Lr2n1(x,d), min(Lr2n1(x,d-1)+P1, delta2)) - delta2;
                    end
                end
                Lr2n(x,d) = L2;
                Sline(((x-1)*D) + d) = Sline(((x-1)*D) + d) + L2;
            end
        end
        minLr2n(x) = min(Lr2n(x,:));    %Find minimum for next delta
        %Sline is incremented. When using indexing, Sline is not updated
        %for each d. Possibly, this must be corrected...
else       
        
        
%% L2, for all d=1...D. Old, correct:        
        for d = 1 : D
            cpd = Crange(d);
            if cpd >= P2
                L2 = P2;
            else
                if d > 1 && d < D
                    %Regular case, d = 2 .. D - 1:
                    L2 = cpd + min(Lr2n1(x,d), min(Lr2n1(x,d+1)+P1, min(Lr2n1(x,d-1)+P1, delta2))) - delta2;
                else
                    %cases d==1 || d==D
                    if d == 1
                       L2 = cpd + min(Lr2n1(x,d), min(Lr2n1(x,d+1)+P1, delta2)) - delta2; 
                    else
                       L2 = cpd + min(Lr2n1(x,d), min(Lr2n1(x,d-1)+P1, delta2)) - delta2;
                    end
                end
            end
            
            Sline(((x-1)*D) + d) = Sline(((x-1)*D) + d) + L2;
            Lr2n(x,d) = L2;
        end
        minLr2n(x) = min(Lr2n(x,:));    %Find minimum for next delta
%%
end        
    end
    
    Lr2n1 = Lr2n;   %Preserve current row for the next run as previous row. Discard previous row. 
    minLr2n1 = minLr2n; %Set current row as previous for next run
    
    %L2 done!
    
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    %L3:
    
    %Dont run to the last x, because L3 comes from x+1
    for x = 1 : relwidth - 1
        Crange = Cline((x-1)*D + 1 : x*D);
        delta3 = minLr3n1(x+1) + P2;
        

if useold ~= 1
        indexrange = indexline((x-1)*D + 1 : x*D);
        Lr3n(x,:) = P2;     %Init Lr3n because not all of its indices are touched any more
        for i = 1 : round(D/2)
            d = indexrange(i);
            if d > 0
                cpd = Crange(d);

                if d > 1 && d < D
                    L3 = cpd + min(Lr3n1(x+1,d), min(Lr3n1(x+1,d+1)+P1, min(Lr3n1(x+1,d-1)+P1, delta3))) - delta3;
                else
                    if d == 1
                        %Check d==1
                        L3 = cpd + min(Lr3n1(x+1,d), min(Lr3n1(x+1,d+1)+P1, delta3)) - delta3;
                    else
                        %check d==D
                        L3 = cpd + min(Lr3n1(x+1,d), min(Lr3n1(x+1,d-1)+P1, delta3)) - delta3;
                    end
                end
                Lr3n(x,d) = L3;    
                Sline(((x-1)*D) + d) = Sline(((x-1)*D) + d) + L3;
            end
        end
        
        minLr3n(x) = min(Lr3n(x,:));
else        
        
%% L3, For all d = 1...D. 
        for d = 1 : D

            cpd = Crange(d);
            if cpd >= P2
                L3 = P2;
            else %end

                if d > 1 && d < D
                    L3 = cpd + min(Lr3n1(x+1,d), min(Lr3n1(x+1,d+1)+P1, min(Lr3n1(x+1,d-1)+P1, delta3))) - delta3;
                else
                    if d == 1
                        %Check d==1
                        L3 = cpd + min(Lr3n1(x+1,d), min(Lr3n1(x+1,d+1)+P1, delta3)) - delta3;
                    else
                        %check d==D
                        L3 = cpd + min(Lr3n1(x+1,d), min(Lr3n1(x+1,d-1)+P1, delta3)) - delta3;
                    end
                end
            end

            Sline(((x-1)*D) + d) = Sline(((x-1)*D) + d) + L3;
            Lr3n(x,d) = L3;
        end
        minLr3n(x) = min(Lr3n(x,:));
%%
end        
        
    end
    
    %Last x for Lr3n:
    Crange = Cline(((relwidth-1)*D)+1 : relwidth * D);  %Get the D costs for the last pixel in the column
    delta3 = 0 + P2;    %delta = minLr3n1(x+1) + P2, but minLr3n1(x+1) doesn't exist at x=relwidth!!!
    

if useold ~= 1    
    %New, indexed:
    Lr3n(relwidth,:) = P2; %Init Lr3n because not all of its indices are touched any more
    indexrange = indexline(((relwidth-1)*D)+1 : relwidth * D);
    for i = 1 : round(D/2)
        d = indexrange(i);    
        if d > 0
            cpd = Crange(d);
            L3 = cpd - delta3;
            Lr3n(relwidth,d) = L3;
        end
    end
    Sline(((x-1)*D) + 1 : ((x-1)*D) + D) = Sline(((x-1)*D) + 1 : ((x-1)*D) + D) + Lr3n(relwidth,:)';
    minLr3n(relwidth) = min(Lr3n(relwidth,:));
    
    Lr3n1 = Lr3n;   
    minLr3n1 = minLr3n;
else


%% L3, last x in line. Old, correct:
    for d = 1 : D
    
        cpd = Crange(d);
        if cpd >= P2
            L3 = P2;
        else %end
            %The regular term doesn't work, because Lr3n1(x+1,...) is not defined,
            %i.e. set to 0. Therefore, it becomes to min(0, 0, 0, delta3) which
            %is 0 because delta3 is P2 is bigger than 0!
            %Lr3n(relwidth,d) = cpd + min(Lr3n1(x+1,d), Lr3n1(x+1,d+1)*P1, Lr3n1(x+1,d-1)+P1), delta3) - delta3;

            %Short cut:
            L3 = cpd - delta3;
            %Lr3n(relwidth,d) = cpd - delta3;
            %Sline((relwidth-1)*D + d) = Sline((relwidth-1)*D + d) + Lr3n(relwidth,d);
        end
        
        Lr3n(relwidth,d) = L3;
        Sline((relwidth-1)*D + d) = Sline((relwidth-1)*D + d) + L3;
    end
    minLr3n(relwidth) = min(Lr3n(relwidth,:));

    Lr3n1 = Lr3n;   
    minLr3n1 = minLr3n;
%%
end

%L3 done
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    

    %Calculate disparity for current line as minimum of aggregated costs:
   
    for x = 1 : relwidth
        currentPix = Sline(((x-1)*D) + 1 : x*D);    %Get relevant (D) aggregated costs for that pixel x
        
        [~,d] = min(currentPix);    %find min of the agg. costs
        
        disparity(y+1,x+D) = d;     %winner take all, i.e. minimum of agg. costs form 
    end
    
    if y == 199
        dispar = disparity(y+1,:);      %#ok<NASGU>. Dumb Matlab doesn't recognize the next statement, where the variable is acutally used!
        save('d:\temp\SGBM_anal\sline_disp.mat','Sline', 'Cline', 'dispar');
        display('Debug data written');
    end
    
    if mod(y-1,100) == 0
        display(['Line ', num2str(y-1), ' processed...']);
    end
    
end     %end of y = 0 : height - 1

save('costpath','Cmatrix', 'indexvolume');

figure(1);
mesh(disparity);

disp('');   %Useful for breakpoints

end %of function
















