--Evaluate how a Bram can be used as LUT. Instantiate a BRAM to be inferred. 
--To force Bram, set ram_style to BRAM in process properties of synthesis 

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

library work;
use work.BramAsLutPack_16x2.ALL;		--Required for the correct constants for Bram component instantiation

ENTITY top is
PORT(
	clk, reset : in std_logic;
	dataout : out unsigned(DataBitSize - 1 downto 0)
);
END top;

ARCHITECTURE Behavioral of top is

COMPONENT bramaslut PORT(
	clk : in std_logic;
	we : in std_logic;		
	waddr : in std_logic_vector(LenInBits - 1 downto 0);	
	wdata : in std_logic_vector(DataBitSize - 1 downto 0);
	raddr : in std_logic_vector(LenInBits - 1 downto 0);
	rdata : out unsigned(DataBitSize - 1 downto 0)
);
END COMPONENT;


signal readaddr, readaddr_next : unsigned(LenInBits - 1 downto 0);

BEGIN

	bram : bramaslut PORT MAP(
		clk => clk,
		we => '0',
		waddr => (OTHERS => '0'), --: in std_logic_vector(3 downto 0);	--0 .. 128, 8 bit
		wdata => (OTHERS => '0'), --: in std_logic_vector(1 downto 0);
		raddr => std_logic_vector(readaddr),
		rdata => dataout
	);
	
	clkProc : PROCESS(clk, reset)
	BEGIN
		IF reset = '1' THEN
			readaddr <= (OTHERS => '0');
		ELSIF clk'event AND clk = '1' THEN
			readaddr <= readaddr_next;
		END IF;
	END PROCESS;
	
	cmbProc : PROCESS(readaddr)
	BEGIN
		readaddr_next <= unsigned(readaddr) + 1;	--4 Bit, wraps around from 1111 to 0000, i.e. from 15 to 0
	END PROCESS;


end Behavioral;

