library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
USE IEEE.math_real.ALL;

library work;
use work.constPack.ALL;

-- TYPE waddrType IS ARRAY(1 TO 20) OF unsigned(7 downto 0);	--1 TO EXPANDLEN of values between 160 and 1
-- TYPE wdataType IS ARRAY(1 TO 20) OF unsigned(9 downto 0);	--1 TO EXPANDLEN of aggregated costs downto 1

-- TYPE raddrType IS ARRAY(1 TO D) OF unsigned(7 downto 0);	--Assign 160 addresses to module and ...
-- TYPE rdataType IS ARRAY(1 TO D) OF unsigned(9 downto 0);	--get 160 data values

ENTITY multififowrapper IS
PORT(
	clktop : in std_logic;
	resettop : in std_logic;
	waddrtop : in waddrType;
	wdatatop : in wdataType;
	--raddrtop : in raddrType;	--Brauchen wir nicht. Ist eh immer 1...160. 
	rdatatop : out rdataType
);
END multififowrapper;

ARCHITECTURE Behavioral of multififowrapper IS

TYPE bufferdaddrType IS ARRAY(1 TO 20) OF unsigned(7 downto 0);	--connect FIFOs	
signal bufferdaddr : bufferdaddrType;

TYPE bufferddataType IS ARRAY(1 TO 20) OF unsigned(9 downto 0);	--connect FIFOs
signal bufferddata : bufferddataType;

signal writeaddress, writeaddress_next : unsigned(ImageWidthInBits - 1 downto 0);
signal readaddress, readaddress_next : unsigned(ImageWidthInBits - 1 downto 0);

signal writeenable : std_logic;

COMPONENT fifoaddr  
	GENERIC (
		DATA : integer := 8;			--use this as data bit width
		ADDRLEN : integer := ImageWidth		--this is the number of elements i.e. ram is able to store ADDRLEN data values!
	);	
	PORT (
		clk : in std_logic;
		we : in std_logic;		--write enable
		waddr : in std_logic_vector(integer(ceil(log2(real(ADDRLEN))))-1 downto 0);	--0 .. 128, 8 bit
		wdata : in std_logic_vector((DATA - 1) downto 0);
		raddr : in std_logic_vector(integer(ceil(log2(real(ADDRLEN))))-1 downto 0);	--0 .. 128, 8 bit
		--rdata : out std_logic_vector((DATA - 1) downto 0)
		rdata : out unsigned((DATA - 1) downto 0)
	);
END COMPONENT;

COMPONENT fifodata
	GENERIC (
		DATA : integer := 10;			--use this as data bit width
		ADDRLEN : integer := ImageWidth		--this is the number of elements i.e. ram is able to store ADDRLEN data values!
	);	
	PORT (
		clk : in std_logic;
		we : in std_logic;		--write enable
		waddr : in std_logic_vector(integer(ceil(log2(real(ADDRLEN))))-1 downto 0);	--0 .. 128, 8 bit
		wdata : in std_logic_vector((DATA - 1) downto 0);
		raddr : in std_logic_vector(integer(ceil(log2(real(ADDRLEN))))-1 downto 0);	--0 .. 128, 8 bit
		rdata : out unsigned((DATA - 1) downto 0)
	);
END COMPONENT;

BEGIN

GENERATE_FIFO_DATA:
	FOR i IN 1 TO EXPANDLEN GENERATE
		myfifo_data : fifodata PORT MAP(
			clk => clktop,
			we => writeenable,
			waddr => std_logic_vector(writeaddress),
			wdata => std_logic_vector(wdatatop(i)),
			raddr => std_logic_vector(readaddress),
			rdata => bufferddata(i)
		);
	END GENERATE GENERATE_FIFO_DATA;
	
GENERATE_FIFO_ADDR:
	FOR i IN 1 TO EXPANDLEN GENERATE
		myfifo_addr : fifoaddr PORT MAP(
			clk => clktop,
			we => writeenable,
			waddr => std_logic_vector(writeaddress),
			wdata => std_logic_vector(waddrtop(i)),
			raddr => std_logic_vector(readaddress),
			rdata => bufferdaddr(i)
		);
	END GENERATE GENERATE_FIFO_ADDR;

	
clkProc : PROCESS(clktop, resettop)
BEGIN
	IF resettop = '1' THEN
		writeenable <= '0';
		writeaddress <= (OTHERS => '0');
		--readaddress(0) <= '1';				--Oder auch readaddress <= to_unsigned(1,ImageWidthInBits);;
		readaddress <= to_unsigned(1,ImageWidthInBits);
		--rdatatop <= (OTHERS => (OTHERS => '0'));
		--rdatasig <= (OTHERS => (OTHERS => '0'));
	ELSIF clktop'event AND clktop = '1' THEN
		writeenable <= '1';
		readaddress <= readaddress_next;
		writeaddress <= writeaddress_next;
		--rdatasig <= rdata_next;
	END IF;
END PROCESS;

--Counters to write/read cycle wise to from FIFOs. writeaddress must always be one ahead...
colcProc : PROCESS(writeaddress, readaddress)
BEGIN
	IF writeaddress < ImageWidth - 1 THEN
		writeaddress_next <= writeaddress + 1;
	ELSE
		writeaddress_next <= (OTHERS => '0');
	END IF;
	
	IF readaddress < ImageWidth - 1 THEN
		readaddress_next <= readaddress + 1;
	ELSE
		readaddress_next <= (OTHERS => '0');
	END IF;
END PROCESS;		


--The 20 address fifos contain numbers between 1 and 160. Goal is to find the indices of the 20 fifos to access with the indices the 20 data fifos
--to get the data values which correspond to the 20 different disparity values between 1 and 160. 
compareProc2 : PROCESS(bufferdaddr, bufferddata)	--ARRAY(1 TO D) OF unsigned(7 downto 0);
CONSTANT P2 : unsigned(9 downto 0) := to_unsigned(144,10);
BEGIN

	rdatatop <= (OTHERS => P2);
	
	--For addresses contained in the ADDR Ram, assign their corresponding data value:
	FOR a in 1 TO 20 LOOP
		IF bufferdaddr(a) > 0 THEN
			rdatatop(to_integer(bufferdaddr(a))) <= bufferddata(a);
		END IF;
	END LOOP;
	
END PROCESS;

--rdatatop <= rdatasig;


END Behavioral;































